from argo import Agent, LLM, Message, Context
import dotenv
import os

from argo.cli import loop


dotenv.load_dotenv()


# Simulate the bank
class Account:
    def __init__(self, balance=0):
        self.balance = balance

    def deposit(self, amount):
        self.balance += amount
        return self.balance

    def withdraw(self, amount):
        if amount > self.balance:
            raise ValueError("Insufficient funds")

        self.balance -= amount
        return self.balance


account = Account(1000)


def callback(chunk: str):
    print(chunk, end="")


agent = Agent(
    name="Banker",
    description="A helpful assistant that can execute bank transactions and reply with the account information.",
    llm=LLM(model=os.getenv("MODEL"), callback=callback, verbose=True),
)


# Add a basic chat skill
@agent.skill
async def casual_chat(ctx: Context) -> Message:
    """Casual chat with the user.

    Use this skill when the user asks a general question or engages
    in casual chat.
    """
    return await ctx.reply()


@agent.skill
async def banker(ctx: Context) -> Message:
    """Interact with the bank account.

    Use this skill when the user asks for information about the bank account,
    such as balance, or asks to deposit or withdraw.
    """
    tool = await ctx.equip()
    result = await ctx.invoke(tool)
    return await ctx.reply(Message.tool(result))


@agent.tool
async def check_balance() -> dict:
    """Returns the balance in the user account.
    """
    return dict(balance=account.balance)


@agent.tool
async def deposit(ammount: int) -> dict:
    """Deposit money into the user account.
    Returns the new balance.
    """
    return dict(balance=account.deposit(ammount), deposited=ammount)


@agent.tool
async def withdraw(ammount: int) -> dict:
    """Withdraw money from the user account.
    Returns the new balance.
    """
    try:
        return dict(balance=account.withdraw(ammount), withdrawn=ammount)
    except:
        return dict(error="Insufficient funds.", balance=account.balance)


loop(agent)



from pydantic import BaseModel
from argo import Agent, LLM, Message, Context
import dotenv
import os
import wikipedia

from argo.cli import loop


dotenv.load_dotenv()


def callback(chunk: str):
    print(chunk, end="")


agent = Agent(
    name="Trivial",
    description="A helpful assistant that can search Wikipedia for answering factual questions.",
    llm=LLM(model=os.getenv("MODEL"), callback=callback, verbose=True),
)


@agent.skill
async def chat(ctx: Context) -> Message:
    """Casual chat with the user.

    Use this only for greetings, basic chat,
    and questions regarding your own capabilities.
    """
    return await ctx.reply()


class Reasoning(BaseModel):
    observation: str
    thought: str
    query: str | None = None
    final: bool


class SearchSummary(BaseModel):
    query: str
    summary: str


@agent.skill
async def question_answering(ctx: Context) -> Message:
    """Answer questions about the world.

    Use this skill when the user asks any questions
    that might require external knowledge.
    """

    ctx.add("You have access to Wikipedia's search engine to answer the user quesstion.")

    for i in range(5):
        reasoning = await ctx.create(
            """
            Breakdown the user request.
            First provide an observation of the
            current state of the task and the knowledge you already have.
            Then, provide a thought on the next step to take.
            Finally, provide a short, concise query for Wikipedia, if necessary.
            Otherwise, if the existing information is enough to answer, set final=True.
            """,
            model=Reasoning
        )

        ctx.add(reasoning)

        if reasoning.final:
            return await ctx.reply()

        results = await ctx.invoke(search)

        summary = await ctx.create(
            """
            Given the desired query, sumarize the fragment of content in the
            following Wikipedia page that is related to the query.
            Make the summary as concise as possible.
            """,
            results,
            model=SearchSummary
        )

        ctx.add(summary)

    return await ctx.reply("Reply with the best available information in the context.")


@agent.tool
async def search(query: str) -> str:
    """Search Wikipedia for information."""
    candidates = wikipedia.search(query, results=1)
    return wikipedia.page(candidates[0]).content


loop(agent)



import abc
import functools
import inspect

from .llm import LLM, Message
from .prompts import DEFAULT_SYSTEM_PROMPT
from .skills import Skill, MethodSkill
from .tools import Tool, MethodTool


class Agent:
    def __init__(
        self,
        name: str,
        description: str,
        llm: LLM,
        *,
        system_prompt=DEFAULT_SYSTEM_PROMPT,
    ):
        self._name = name
        self._description = description
        self._llm = llm
        self._skills = []
        self._tools = []
        self._system_prompt = system_prompt.format(name=name, description=description)

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._description

    @property
    def llm(self):
        return self._llm

    async def perform(self, messages: list[Message]) -> Message:
        from .context import Context
        """Main entrypoint for the agent.

        This method will select the right skill to perform the task and then execute it.
        The skill is selected based on the messages and the skills available to the agent.
        """
        context = Context(self, [Message.system(self._system_prompt)] + list(messages))
        skill = await context.engage()
        return await skill.execute(context)

    def skill(self, target):
        if isinstance(target, Skill):
            self._skills.append(target)
            return target

        if not callable(target):
            raise ValueError("Skill must be a callable.")

        if not inspect.iscoroutinefunction(target):
            raise ValueError("Skill must be a coroutine function.")

        name = target.__name__
        description = inspect.getdoc(target)
        skill = MethodSkill(name, description, target)
        self._skills.append(skill)
        return skill

    def tool(self, target):
        if isinstance(target, Tool):
            self._tools.append(target)
            return target

        # BUG: Doesn't work for sync method
        if not inspect.iscoroutinefunction(target):

            @functools.wraps(target)
            async def wrapper(*args, **kwargs):
                return target(*args, **kwargs)

            target = wrapper

        name = target.__name__
        description = inspect.getdoc(target)
        tool = MethodTool(name, description, target)
        self._tools.append(tool)
        return tool



import inspect
import abc


class Tool:
    def __init__(self, name: str, description: str):
        self._name = name
        self._description = description

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._description

    @abc.abstractmethod
    def parameters(self) -> dict[str, type]:
        pass

    @abc.abstractmethod
    async def run(self, **kwargs):
        pass


class MethodTool(Tool):
    def __init__(self, name, description, target):
        super().__init__(name, description)
        self._target = target

    def parameters(self):
        args = inspect.get_annotations(self._target)
        return {name: type for name, type in args.items() if name != "return"}

    async def run(self, **kwargs):
        return await self._target(**kwargs)



import asyncio
import sys

from pathlib import Path
import dotenv
import rich
import rich.prompt
from .agent import Agent
from .llm import LLM, Message
from .declarative import parse


def loop(agent:Agent):
    """Runs a CLI agent loop with integrated
    conversation history management.

    This method creates an async context internally,
    and handles storing and retrieving conversation history.
    It also handles keyboard interrupts and EOF errors.

    This method blocks the terminal waiting for user input,
    and loops until EOF (Ctrl+D) is pressed.
    """
    rich.print(f"[bold green]{agent.name}[/bold green]: {agent.description}\n")

    if agent.llm.verbose:
        rich.print(f"[yellow]Running in verbose mode.[/yellow]")

    rich.print(f"[yellow]Press Ctrl+D to exit at any time.\n[/yellow]")

    async def run():
        history = []

        while True:
            try:
                user_input = input(">>> ")
                history.append(Message.user(user_input))
                response = await agent.perform(history)
                history.append(response)
                print("\n")
            except (EOFError, KeyboardInterrupt):
                break

    asyncio.run(run())


def main():
    dotenv.load_dotenv()
    import os

    path = Path(sys.argv[1])

    API_KEY = os.getenv("API_KEY")
    BASE_URL = os.getenv("BASE_URL")
    MODEL = os.getenv("MODEL")

    def callback(chunk: str):
        print(chunk, end="")

    llm = LLM(model=MODEL, api_key=API_KEY, base_url=BASE_URL, callback=callback)

    config = parse(path)
    # rich.print(config)
    # rich.print()

    agent = config.compile(llm)
    loop(agent)


if __name__ == "__main__":
    main()



from argo import Agent, LLM, Message, Context
from argo.cli import loop
import dotenv
import os


dotenv.load_dotenv()


def callback(chunk:str):
    print(chunk, end="")


agent = Agent(
    name="Agent",
    description="A helpful assistant.",
    llm=LLM(model=os.getenv("MODEL"), callback=callback, verbose=True),
)


@agent.skill
async def chat(ctx: Context) -> Message:
    """Casual chat with the user.
    """
    return await ctx.reply()


loop(agent)



import inspect
from typing import Callable, Type, TypeVar
import rich
import json
import openai
from pydantic import BaseModel
import os


class Message(BaseModel):
    role: str
    content: str

    @classmethod
    def system(cls, content: str) -> "Message":
        return cls(role="system", content=content)

    @classmethod
    def user(cls, content: str) -> "Message":
        return cls(role="user", content=content)

    @classmethod
    def assistant(cls, content: str) -> "Message":
        return cls(role="assistant", content=content)

    @classmethod
    def tool(cls, content: BaseModel) -> "Message":
        return cls(role="tool", content=content.model_dump_json())


T = TypeVar("T", bound=BaseModel)

LLMCallback = Callable[[str], None]


class LLM:
    def __init__(
        self,
        model: str,
        callback: LLMCallback = None,
        verbose: bool = False,
        base_url: str = None,
        api_key: str = None,
    ):
        self.model = model
        self.verbose = verbose

        if base_url is None:
            base_url = os.getenv("BASE_URL")
        if api_key is None:
            api_key = os.getenv("API_KEY")

        self.client = openai.AsyncOpenAI(base_url=base_url, api_key=api_key)
        self.callback = callback

    async def chat(self, messages: list[Message], **kwargs) -> Message:
        result = []

        async for chunk in await self.client.chat.completions.create(
            model=self.model,
            messages=[message.model_dump() for message in messages],
            stream=True,
            **kwargs,
        ):
            content = chunk.choices[0].delta.content

            if content is None:
                continue

            if self.callback:
                if inspect.iscoroutinefunction(self.callback):
                    await self.callback(content)
                else:
                    self.callback(content)

            result.append(content)

        return Message.assistant("".join(result))

    async def parse(self, model: Type[T], messages: list[Message], **kwargs) -> T:
        response = await self.client.beta.chat.completions.parse(
            model=self.model,
            messages=[message.model_dump() for message in messages],
            response_format=model,
            **kwargs,
        )

        result = response.choices[0].message.parsed

        if self.verbose:
            rich.print(result)

        return result



import json
from typing import TypeVar
from pydantic import BaseModel, create_model
import rich

from .agent import Agent
from .llm import Message
from .prompts import *
from .skills import Skill
from .tools import Tool
from .utils import generate_pydantic_code


class Choose(BaseModel):
    reasoning: str
    selection: str


class Decide(BaseModel):
    reasoning: str
    answer: bool


class ToolResult(BaseModel):
    tool: str
    description: str
    result: str


class Equip(BaseModel):
    reasoning: str
    tool: str


class Engage(BaseModel):
    reasoning: str
    skill: str


T = TypeVar("T")
TModel = TypeVar("TModel", bound=BaseModel)


class Context:
    def __init__(self, agent: Agent, messages: list[Message]):
        self.agent = agent
        self._messages = messages

    @property
    def messages(self) -> list[Message]:
        return list(self._messages)

    def _wrap(self, message: Message | str | BaseModel):
        if isinstance(message, Message):
            return message
        elif isinstance(message, str):
            return Message.system(message)
        elif isinstance(message, BaseModel):
            return Message.tool(message)

        raise TypeError(f"Invalid message type: {type(message)}")

    def _expand_content(self, *instructions):
        messages = self.messages

        for message in instructions:
            messages.append(self._wrap(message))

        return messages

    async def reply(
        self, *instructions: str | Message, persistent: bool = False
    ) -> Message:
        """Reply to the provided messages.

        This method will use the LLM to generate a response to the provided messages.
        It does not use any skills.
        Mostly useful inside skills to finish the conversation.
        """
        result = await self.agent.llm.chat(self._expand_content(*instructions))

        if persistent:
            self.add(result)

        return result

    async def choose(self, options: list[T], *instructions: str | Message) -> T:
        """Choose one option out of many.

        This method will use the LLM to choose one option out of many.
        It does not use any skills.
        Mostly useful inside skills to make decisions.
        """
        mapping = {str(option): option for option in options}

        prompt = DEFAULT_CHOOSE_PROMPT.format(
            options="\n".join([f"- {option}" for option in options]),
            format=Choose.model_json_schema(),
        )

        response = await self.agent.llm.parse(
            Choose, self._expand_content(*instructions, Message.system(prompt))
        )

        return mapping[response.selection]

    async def decide(self, *instructions) -> bool:
        """Decide True or False.

        This method will use the LLM to decide True or False.
        It does not use any skills.
        Mostly useful inside skills to make decisions.
        """
        prompt = DEFAULT_DECIDE_PROMPT.format(
            format=Decide.model_json_schema(),
        )

        response = await self.agent.llm.parse(
            Decide, self._expand_content(*instructions, Message.system(prompt))
        )

        return response.answer

    async def equip(
        self, *instructions: str | Message, tools: list[Tool] = None
    ) -> Tool:
        """Selects one and exactly one tool.

        This method will use the LLM to pick a tool from the list of tools.
        It does not use any skills.
        Mostly useful inside skills to make decisions.
        """
        if tools is None:
            tools = self.agent._tools

        tool_str = {tool.name: tool.description for tool in tools}
        mapping = {tool.name: tool for tool in tools}

        prompt = DEFAULT_EQUIP_PROMPT.format(
            tools=tool_str,
            format=Equip.model_json_schema(),
        )

        response = await self.agent.llm.parse(
            Equip, self._expand_content(*instructions, Message.system(prompt))
        )

        return mapping[response.selection]

    async def engage(self, *instructions: str | Message) -> Skill:
        """
        Selects a single skill to respond to the instructions.
        This method will use the LLM to pick a skill from the list of skills.
        """
        skills: list[Skill] = self.agent._skills
        skills_map = {s.name: s for s in skills}

        prompt = DEFAULT_ENGAGE_PROMPT.format(
            skills="\n".join(
                [f"- {skill.name}: {skill.description}" for skill in skills]
            ),
            format=Engage.model_json_schema(),
        )

        messages = self._expand_content(*instructions, Message.system(prompt))

        response = await self.agent.llm.parse(Engage, messages)
        return skills_map[response.skill]

    async def invoke(
        self, tool: Tool = None, *instructions: str | Message, **kwargs
    ) -> ToolResult:
        """
        Invokes a tool with the given instructions.
        This method will use the LLM to generate the parameters for the tool.
        The tool will then be invoked with the generated parameters.
        """
        if tool is None:
            tool = await self.equip(instructions)

        parameters = tool.parameters()

        for k, v in kwargs.items():
            parameters[k] = (parameters[k], v)

        tool_name_camel_case = tool.name.title().replace("_", "")
        model_cls: type[BaseModel] = create_model(tool_name_camel_case, **parameters)

        prompt = DEFAULT_INVOKE_PROMPT.format(
            name=tool.name,
            defaults=kwargs,
            parameters={k: v for k, v in parameters.items() if k not in kwargs},
            description=tool.description,
            format=model_cls.model_json_schema(),
        )

        messages = self._expand_content(*instructions, Message.system(prompt))

        response: BaseModel = await self.agent.llm.parse(
            model_cls, messages + [Message.system(prompt)]
        )

        return ToolResult(
            tool=tool.name,
            description=tool.description,
            result=str(await tool.run(**response.model_dump())),
        )

    async def create(self, *instructions: str | Message, model: type[TModel]) -> TModel:
        """
        Parses the given instructions into a model.
        This method will use the LLM to generate the parameters for the model.
        """
        messages = self._expand_content(*instructions)

        model_code = generate_pydantic_code(model)

        prompt = Message.system(DEFAULT_CREATE_PROMPT.format(
            type=model.__name__,
            signature=model_code,
            docs=model.__doc__ or "",
            format=json.dumps(model.model_json_schema(), indent=2),
        ))

        return await self.agent.llm.parse(model, messages)

    def add(self, *messages: Message | str | BaseModel) -> None:
        """
        Appends a message to the context.
        """
        for message in messages:
            self._messages.append(self._wrap(message))



import abc
from .llm import Message


class Skill:
    def __init__(self, name: str, description: str):
        self._name = name
        self._description = description
        self._before = []

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._description

    @abc.abstractmethod
    async def _execute(self, ctx) -> Message:
        pass

    async def execute(self, ctx) -> Message:
        messages = list(ctx.messages)

        for skill in self._before:
            m = await skill.execute(ctx)
            messages.append(m)

        return await self._execute(ctx)

    def requires(self, target):
        if isinstance(target, Skill):
            self._before.append(target)
        else:
            self._before.append(MethodSkill(target.__name__, target.__doc__, target))

        return target


class MethodSkill(Skill):
    def __init__(self, name: str, description: str, target):
        super().__init__(name, description)
        self._target = target

    async def _execute(self, ctx) -> Message:
        return await self._target(ctx)
